@@ -7,7 +7,6 @@ import copy
 import os
 import re
 import sys
-from code import compile_command
 try:
     from cStringIO import StringIO
 except ImportError:
@@ -32,7 +31,7 @@ except ImportError:
     pep8 = None
 
 
-__version__ = '0.6.3'
+__version__ = '0.6.4'
 
 
 PEP8_BIN = 'pep8'
@@ -40,6 +39,7 @@ PEP8_PASSES_MAX = 100
 CR = '\r'
 LF = '\n'
 CRLF = '\r\n'
+MAX_LINE_WIDTH = 79
 
 
 def open_with_encoding(filename, encoding, mode='r'):
@@ -55,8 +55,17 @@ def detect_encoding(filename):
     """Return file encoding."""
     try:
         # Python 3
-        with open(filename, 'rb') as input_file:
-            return tokenize.detect_encoding(input_file.readline)[0]
+        try:
+            with open(filename, 'rb') as input_file:
+                encoding = tokenize.detect_encoding(input_file.readline)[0]
+
+            # Check for correctness of encoding
+            with open(filename, encoding=encoding) as input_file:
+                input_file.read()
+
+            return encoding
+        except (SyntaxError, LookupError, UnicodeDecodeError):
+            return 'latin-1'
     except AttributeError:
         return 'utf-8'
 
@@ -109,9 +118,13 @@ class FixPEP8(object):
         self.indent_word = _get_indentword("".join(self.source))
         # method definition
         self.fix_e111 = self.fix_e101
-        self.fix_e222 = self.fix_e221
-        self.fix_e223 = self.fix_e221
-        self.fix_e241 = self.fix_e221
+        self.fix_e202 = self.fix_e201
+        self.fix_e203 = self.fix_e201
+        self.fix_e211 = self.fix_e201
+        self.fix_e221 = self.fix_e271
+        self.fix_e222 = self.fix_e271
+        self.fix_e223 = self.fix_e271
+        self.fix_e241 = self.fix_e271
         self.fix_e242 = self.fix_e224
         self.fix_e261 = self.fix_e262
         self.fix_e272 = self.fix_e271
@@ -141,24 +154,19 @@ class FixPEP8(object):
                 elif modified_lines == []:  # Empty list means no fix
                     if self.options.verbose:
                         sys.stderr.write('Not fixing {f} on line {l}\n'.format(
-                                f=result['id'], l=result['line']))
+                            f=result['id'], l=result['line']))
                 else:  # We assume one-line fix when None
                     completed_lines.append(result['line'])
             else:
                 if self.options.verbose:
                     sys.stderr.write("'%s' is not defined.\n" %
-                            fixed_methodname)
+                                     fixed_methodname)
                     info = result['info'].strip()
                     sys.stderr.write("%s:%s:%s:%s\n" % (self.filename,
                                                         result['line'],
                                                         result['column'],
                                                         info))
 
-    def _fix_whitespace(self, result, pattern, repl):
-        (indent, rest) = _split_indentation(self.source[result['line'] - 1])
-        fixed = re.sub(pattern, repl, rest)
-        self.source[result['line'] - 1] = indent + fixed
-
     def fix(self):
         pep8_options = self._pep8_options(self.filename)
         if pep8:
@@ -179,33 +187,13 @@ class FixPEP8(object):
             return []
 
     def fix_e201(self, result):
-        self._fix_whitespace(result, r"\(\s+", "(")
-        self._fix_whitespace(result, r"\[\s+", "[")
-        self._fix_whitespace(result, r"{\s+", "{")
+        line_index = result['line'] - 1
+        target = self.source[line_index]
+        offset = result['column'] - 1
 
-    def fix_e202(self, result):
-        self._fix_whitespace(result, r"\s+\)", ")")
-        self._fix_whitespace(result, r"\s+\]", "]")
-        self._fix_whitespace(result, r"\s+}", "}")
-
-    def fix_e203(self, result):
-        self._fix_whitespace(result, r"\s+:", ":")
-        self._fix_whitespace(result, r"\s+,", ",")
-
-    def fix_e211(self, result):
-        self._fix_whitespace(result, r"\s+\(", "(")
-        self._fix_whitespace(result, r"\s+\[", "[")
-
-    def fix_e221(self, result):
-        """Fix E221, E222, and E223."""
-        target = self.source[result['line'] - 1]
-        c = result['column'] + 1
-        fixed = re.sub(r'\s+', ' ', target[c::-1], 1)[::-1] + target[c + 1:]
-        if fixed == target:
-            # for e223 fixed method
-            fixed = re.sub(r'\t+', ' ', target[c::-1], 1)[::-1] + \
-                    target[c + 1:]
-        self.source[result['line'] - 1] = fixed
+        self.source[line_index] = fix_whitespace(target,
+                                                 offset=offset,
+                                                 replacement='')
 
     def fix_e224(self, result):
         target = self.source[result['line'] - 1]
@@ -222,7 +210,7 @@ class FixPEP8(object):
         # Only proceed if non-whitespace characters match.
         # And make sure we don't break the indentation.
         if (fixed.replace(' ', '') == target.replace(' ', '') and
-            _get_indentation(fixed) == _get_indentation(target)):
+                _get_indentation(fixed) == _get_indentation(target)):
             self.source[result['line'] - 1] = fixed
         else:
             return []
@@ -249,11 +237,11 @@ class FixPEP8(object):
         #     def foo(a=\
         #             1)
         if (fixed.endswith('=\\\n') or
-            fixed.endswith('=\\\r\n') or
-            fixed.endswith('=\\\r')):
+                fixed.endswith('=\\\r\n') or
+                fixed.endswith('=\\\r')):
             self.source[line_index] = fixed.rstrip('\n\r \t\\')
             self.source[line_index + 1] = \
-                    self.source[line_index + 1].lstrip()
+                self.source[line_index + 1].lstrip()
             return [line_index + 1, line_index + 2]  # Line indexed at 1
 
         self.source[result['line'] - 1] = fixed
@@ -277,12 +265,9 @@ class FixPEP8(object):
         target = self.source[line_index]
         offset = result['column'] - 1
 
-        # Proceed safely
-        if target[offset] not in [' ', '\t']:
-            return []
-
-        fixed = target[:offset].rstrip() + ' ' + target[offset:].lstrip()
-        self.source[line_index] = fixed
+        self.source[line_index] = fix_whitespace(target,
+                                                 offset=offset,
+                                                 replacement=' ')
 
     def fix_e301(self, result):
         cr = self.newline
@@ -325,41 +310,55 @@ class FixPEP8(object):
         indentation = target.split("import ")[0]
         modules = target.split("import ")[1].split(",")
         fixed_modulelist = \
-                [indentation + "import %s" % m.lstrip() for m in modules]
+            [indentation + "import %s" % m.lstrip() for m in modules]
         self.source[line_index] = self.newline.join(fixed_modulelist)
 
     def fix_e501(self, result):
-        # FIXME: lazy implementation
-        RETURN_COLUMN = 75
         line_index = result['line'] - 1
         target = self.source[line_index]
-        indent = _get_indentation(target)
-        source = target[len(indent):]
-        sio = StringIO(target)
 
-        # don't fix when multiline string
-        try:
-            tokens = tokenize.generate_tokens(sio.readline)
-            _tokens = [t for t in tokens]
-        except (tokenize.TokenError, IndentationError):
+        if target.lstrip().startswith('#'):
+            # Wrap commented lines
+            self.source[line_index] = shorten_comment(line=target,
+                                                      newline=self.newline)
             return
+        else:
+            # FIXME: lazy implementation
+            #RETURN_COLUMN = 75
+            indent = _get_indentation(target)
+            source = target[len(indent):]
+            sio = StringIO(target)
 
-        # line sparate with OPERATOR
-        _tokens.reverse()
-        for tkn in _tokens:
-            if token.OP == tkn[0]:
-                offset = tkn[2][1] + 1
-                if offset > (79 - len(indent) - 4):
-                    continue
-                fixed = "%s" % source[:offset - len(indent)] + "\n" + \
-                        indent + "    " + source[offset - len(indent):]
-                try:
-                    ret = compile_command(fixed)
-                except SyntaxError:
-                    continue
-                if ret:
-                    self.source[line_index] = indent + fixed
-                    return
+            # don't fix when multiline string
+            try:
+                tokens = tokenize.generate_tokens(sio.readline)
+                _tokens = [t for t in tokens]
+            except (tokenize.TokenError, IndentationError):
+                return []
+
+            # Prefer
+            # my_long_function_name(
+            #     x, y, z, ...)
+            #
+            # over
+            # my_long_function_name(x, y,
+            #     z, ...)
+            candidate0 = _shorten_line(_tokens, source, target, indent,
+                                       self.indent_word, reverse=False)
+            candidate1 = _shorten_line(_tokens, source, target, indent,
+                                       self.indent_word, reverse=True)
+            if candidate0 and candidate1:
+                if candidate0.split(self.newline)[0].endswith('('):
+                    self.source[line_index] = candidate0
+                else:
+                    self.source[line_index] = candidate1
+            elif candidate0:
+                self.source[line_index] = candidate0
+            elif candidate1:
+                self.source[line_index] = candidate1
+            else:
+                # Otherwise both don't work
+                return []
 
         # FIXME: disable now
         #for offset in range(50):
@@ -385,14 +384,19 @@ class FixPEP8(object):
         target = self.source[line_index]
         c = result['column']
 
-        fixed_source = target[:c] + self.newline + \
-                       _get_indentation(target) + self.indent_word + \
-                       target[c:].lstrip()
+        fixed_source = (target[:c] + self.newline +
+                        _get_indentation(target) + self.indent_word +
+                        target[c:].lstrip())
         self.source[result['line'] - 1] = fixed_source
 
     def fix_e702(self, result):
         """Fix multiple statements on one line."""
-        target = self.source[result['line'] - 1]
+        line_index = result['line'] - 1
+        target = self.source[line_index]
+
+        if target.rstrip().endswith(';'):
+            self.source[line_index] = target.rstrip('\n \r\t;') + self.newline
+            return
 
         # We currently do not support things like
         #     """
@@ -412,7 +416,7 @@ class FixPEP8(object):
         f = [_get_indentation(target) + t.strip()
              for t in [first, second] if t.strip()]
 
-        self.source[result['line'] - 1] = self.newline.join(f) + self.newline
+        self.source[line_index] = self.newline.join(f) + self.newline
 
     def fix_w291(self, result):
         fixed_line = self.source[result['line'] - 1].rstrip()
@@ -504,11 +508,11 @@ class FixPEP8(object):
             return []
 
         if (line.endswith('\\\n') or
-            line.endswith('\\\r\n') or
-            line.endswith('\\\r')):
+                line.endswith('\\\r\n') or
+                line.endswith('\\\r')):
             self.source[line_index] = line.rstrip('\n\r \t\\')
             self.source[line_index + 1] = \
-                    ' ' + self.source[line_index + 1].lstrip()
+                ' ' + self.source[line_index + 1].lstrip()
             return [line_index + 1, line_index + 2]  # Line indexed at 1
 
         modified_lines = [1 + line_index]  # Line indexed at 1
@@ -536,7 +540,7 @@ class FixPEP8(object):
 
             # We do not handle anything other than plain multiline strings
             if ('(' in self.source[end_line_index] or
-                '\\' in self.source[end_line_index]):
+                    '\\' in self.source[end_line_index]):
                 return []
 
             for i in range(line_index + 1, end_line_index + 1):
@@ -616,7 +620,7 @@ class FixPEP8(object):
         start = target.find('`')
         end = target[::-1].find('`') * -1
         self.source[result['line'] - 1] = "%srepr(%s)%s" % (
-                target[:start], target[start + 1:end - 1], target[end:])
+            target[:start], target[start + 1:end - 1], target[end:])
 
 
 def _find_newline(source):
@@ -647,7 +651,7 @@ def _get_indentword(source):
             if t[0] == token.INDENT:
                 indent_word = t[1]
                 break
-    except tokenize.TokenError:
+    except (tokenize.TokenError, IndentationError):
         pass
     return indent_word
 
@@ -687,8 +691,8 @@ def _priority_key(pep8_result):
 
     """
     priority = ['e101', 'e111', 'w191',  # Global fixes
-                'e702',  # Break multiline statements early
-               ]
+                'e701',  # Fix multiline colon-based before semicolon based
+                'e702']  # Break multiline statements early
     key = pep8_result['id'].lower()
     if key in priority:
         return priority.index(key)
@@ -736,6 +740,68 @@ def _fix_basic_raise(line, newline):
                     comment, newline])
 
 
+def _shorten_line(tokens, source, target, indentation, indent_word,
+                  reverse=False):
+    """Separate line at OPERATOR."""
+    if reverse:
+        tokens.reverse()
+    for tkn in tokens:
+        # Don't break on '=' after keyword as this violates PEP 8.
+        if token.OP == tkn[0] and tkn[1] != '=':
+            offset = tkn[2][1] + 1
+            if reverse:
+                if offset > (MAX_LINE_WIDTH - len(indentation) -
+                             len(indent_word)):
+                    continue
+            else:
+                if (len(target.rstrip()) - offset >
+                        (MAX_LINE_WIDTH - len(indentation) -
+                         len(indent_word))):
+                    continue
+            first = source[:offset - len(indentation)]
+            second = (indentation + indent_word +
+                      source[offset - len(indentation):])
+            if not second.strip():
+                continue
+            # Don't modify if lines are not short enough
+            if len(first) > MAX_LINE_WIDTH or len(second) > MAX_LINE_WIDTH:
+                continue
+            fixed = first + "\n" + second
+            try:
+                ret = compile(fixed, '<string>', 'exec')
+            except SyntaxError:
+                continue
+            if ret:
+                return indentation + fixed
+    return None
+
+
+def shorten_comment(line, newline):
+    """Return trimmed or split long comment line."""
+    assert len(line) > MAX_LINE_WIDTH
+    line = line.rstrip()
+
+    MIN_CHARACTER_REPEAT = 5
+    if len(line) - len(line.rstrip(line[-1])) >= MIN_CHARACTER_REPEAT:
+        # Trim comments that end with things like ---------
+        return line[:MAX_LINE_WIDTH] + newline
+    else:
+        indentation = _get_indentation(line) + '# '
+        import textwrap
+        split_lines = textwrap.wrap(line.lstrip(' \t#'),
+                                    initial_indent=indentation,
+                                    subsequent_indent=indentation,
+                                    width=MAX_LINE_WIDTH)
+        return newline.join(split_lines) + newline
+
+
+def fix_whitespace(line, offset, replacement):
+    """Replace whitespace at offset and return fixed line."""
+    # Replace escaped newlines too
+    return (line[:offset].rstrip('\n\r \t\\') +
+            replacement + line[offset:].lstrip('\n\r \t\\'))
+
+
 def _spawn_pep8(pep8_options):
     """Execute pep8 via subprocess.Popen."""
     paths = os.environ['PATH'].split(':')
@@ -744,14 +810,15 @@ def _spawn_pep8(pep8_options):
             cmd = ([os.path.join(path, PEP8_BIN)] +
                    pep8_options)
             p = Popen(cmd, stdout=PIPE)
-            return [_analyze_pep8result(l.decode('utf8'))
-                    for l in p.stdout.readlines()]
+            output = p.communicate()[0].decode('utf-8')
+            return [_analyze_pep8result(l)
+                    for l in output.splitlines()]
     raise Exception("'%s' is not found." % PEP8_BIN)
 
 
 def _execute_pep8(pep8_options, source):
     """Execute pep8 via python method calls."""
-    pep8.options, pep8.args = (pep8.process_options(['pep8'] + pep8_options))
+    pep8.process_options(['pep8'] + pep8_options)
 
     class QuietChecker(pep8.Checker):
 
@@ -763,7 +830,9 @@ def _execute_pep8(pep8_options, source):
 
         def report_error(self, line_number, offset, text, check):
             """Collect errors."""
-            self.__results.append(
+            code = text[:4]
+            if not pep8.ignore_code(code):
+                self.__results.append(
                     dict(id=text.split()[0], line=line_number,
                          column=offset + 1, info=text))
 
@@ -812,7 +881,7 @@ class Reindenter(object):
         try:
             for t in tokens:
                 self.tokeneater(*t)
-        except IndentationError:
+        except (tokenize.TokenError, IndentationError):
             return False
         # Remove trailing empty lines.
         lines = self.lines
@@ -856,8 +925,8 @@ class Reindenter(object):
                         for j in range(i - 1, -1, -1):
                             jline, jlevel = stats[j]
                             if jlevel >= 0:
-                                want = have + _getlspace(after[jline - 1]) - \
-                                       _getlspace(lines[jline])
+                                want = (have + _getlspace(after[jline - 1]) -
+                                        _getlspace(lines[jline]))
                                 break
                     if want < 0:
                         # Still no luck -- leave it alone.
@@ -992,7 +1061,7 @@ def parse_args(args):
     parser.add_option('-p', '--pep8-passes',
                       default=PEP8_PASSES_MAX, type='int',
                       help='maximum number of additional pep8 passes'
-                           ' (default:%d)' % PEP8_PASSES_MAX)
+                           ' (default: %default)')
     parser.add_option('--ignore', default='',
                       help='do not fix these errors/warnings (e.g. E4,W)')
     parser.add_option('--select', default='',
